/*

Copyright (c) 2014-2017 Alex Forencich

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

*/

// Language: Verilog 2001

`timescale 1ns / 1ps

/*
 * AXI4-Stream asynchronous FIFO
 */
module axis_async_fifo #
(
    parameter DEPTH_LOG = 10,
    parameter TDATA_SIZE = 16,
    parameter TUSER_SIZE = 2
)
(
    /*
     * Common asynchronous reset
     */
    input  wire                   aresetn,

    /*
     * AXI Stream input
     */
    input  wire                   s_axis_aclk,
    input  wire [TDATA_SIZE-1:0]  s_axis_tdata,
    input  wire                   s_axis_tvalid,
    output wire                   s_axis_tready,
    input  wire                   s_axis_tlast,
    input  wire [TUSER_SIZE-1:0]  s_axis_tuser,
    
    /*
     * AXI Stream output
     */
    input  wire                   m_axis_aclk,
    output wire [TDATA_SIZE-1:0]  m_axis_tdata,
    output wire                   m_axis_tvalid,
    input  wire                   m_axis_tready,
    output wire                   m_axis_tlast,
    output wire [TUSER_SIZE-1:0]  m_axis_tuser
);

reg [DEPTH_LOG:0] wr_ptr_reg = {DEPTH_LOG+1{1'b0}}, wr_ptr_next;
reg [DEPTH_LOG:0] wr_ptr_gray_reg = {DEPTH_LOG+1{1'b0}}, wr_ptr_gray_next;
reg [DEPTH_LOG:0] wr_addr_reg = {DEPTH_LOG+1{1'b0}};
reg [DEPTH_LOG:0] rd_ptr_reg = {DEPTH_LOG+1{1'b0}}, rd_ptr_next;
reg [DEPTH_LOG:0] rd_ptr_gray_reg = {DEPTH_LOG+1{1'b0}}, rd_ptr_gray_next;
reg [DEPTH_LOG:0] rd_addr_reg = {DEPTH_LOG+1{1'b0}};

(* ASYNC_REG = "TRUE" *) reg [DEPTH_LOG:0] wr_ptr_gray_sync1_reg = {DEPTH_LOG+1{1'b0}};
(* ASYNC_REG = "TRUE" *) reg [DEPTH_LOG:0] wr_ptr_gray_sync2_reg = {DEPTH_LOG+1{1'b0}};
(* ASYNC_REG = "TRUE" *) reg [DEPTH_LOG:0] rd_ptr_gray_sync1_reg = {DEPTH_LOG+1{1'b0}};
(* ASYNC_REG = "TRUE" *) reg [DEPTH_LOG:0] rd_ptr_gray_sync2_reg = {DEPTH_LOG+1{1'b0}};

reg input_rst_sync1_reg = 1'b1;
reg input_rst_sync2_reg = 1'b1;
reg input_rst_sync3_reg = 1'b1;
reg output_rst_sync1_reg = 1'b1;
reg output_rst_sync2_reg = 1'b1;
reg output_rst_sync3_reg = 1'b1;

localparam MEM_WORD_SIZE = TDATA_SIZE+1+TUSER_SIZE;

reg [MEM_WORD_SIZE-1:0] mem[(2**DEPTH_LOG)-1:0];
reg [MEM_WORD_SIZE-1:0] mem_read_data_reg = {MEM_WORD_SIZE{1'b0}};
reg mem_read_data_valid_reg = 1'b0, mem_read_data_valid_next;
wire [MEM_WORD_SIZE-1:0] mem_write_data;

reg [MEM_WORD_SIZE-1:0] output_data_reg = {MEM_WORD_SIZE{1'b0}};

reg m_axis_tvalid_reg = 1'b0, m_axis_tvalid_next;

// full when first TWO MSBs do NOT match, but rest matches
// (gray code equivalent of first MSB different but rest same)
wire full = ((wr_ptr_gray_reg[DEPTH_LOG] != rd_ptr_gray_sync2_reg[DEPTH_LOG]) &&
             (wr_ptr_gray_reg[DEPTH_LOG-1] != rd_ptr_gray_sync2_reg[DEPTH_LOG-1]) &&
             (wr_ptr_gray_reg[DEPTH_LOG-2:0] == rd_ptr_gray_sync2_reg[DEPTH_LOG-2:0]));
// empty when pointers match exactly
wire empty = rd_ptr_gray_reg == wr_ptr_gray_sync2_reg;

// control signals
reg write;
reg read;
reg store_output;

assign s_axis_tready = ~full & ~input_rst_sync3_reg;

assign m_axis_tvalid = m_axis_tvalid_reg;

assign mem_write_data = {s_axis_tlast, s_axis_tuser, s_axis_tdata};
assign {m_axis_tlast, m_axis_tuser, m_axis_tdata} = output_data_reg;

// reset synchronization
always @(posedge s_axis_aclk or negedge aresetn) begin
    if (~aresetn) begin
        input_rst_sync1_reg <= 1'b1;
        input_rst_sync2_reg <= 1'b1;
        input_rst_sync3_reg <= 1'b1;
    end else begin
        input_rst_sync1_reg <= 1'b0;
        input_rst_sync2_reg <= input_rst_sync1_reg | output_rst_sync1_reg;
        input_rst_sync3_reg <= input_rst_sync2_reg;
    end
end

always @(posedge m_axis_aclk or negedge aresetn) begin
    if (~aresetn) begin
        output_rst_sync1_reg <= 1'b1;
        output_rst_sync2_reg <= 1'b1;
        output_rst_sync3_reg <= 1'b1;
    end else begin
        output_rst_sync1_reg <= 1'b0;
        output_rst_sync2_reg <= input_rst_sync1_reg | output_rst_sync1_reg;
        output_rst_sync3_reg <= output_rst_sync2_reg;
    end
end

// Write logic
always @* begin
    write = 1'b0;

    wr_ptr_next = wr_ptr_reg;
    wr_ptr_gray_next = wr_ptr_gray_reg;

    if (s_axis_tvalid) begin
        // input data valid
        if (~full) begin
            // not full, perform write
            write = 1'b1;
            wr_ptr_next = wr_ptr_reg + 1;
            wr_ptr_gray_next = wr_ptr_next ^ (wr_ptr_next >> 1);
        end
    end
end

always @(posedge s_axis_aclk) begin
    if (input_rst_sync3_reg) begin
        wr_ptr_reg <= {DEPTH_LOG+1{1'b0}};
        wr_ptr_gray_reg <= {DEPTH_LOG+1{1'b0}};
    end else begin
        wr_ptr_reg <= wr_ptr_next;
        wr_ptr_gray_reg <= wr_ptr_gray_next;
    end

    wr_addr_reg <= wr_ptr_next;

    if (write) begin
        mem[wr_addr_reg[DEPTH_LOG-1:0]] <= mem_write_data;
    end
end

// pointer synchronization
always @(posedge s_axis_aclk) begin
    if (input_rst_sync3_reg) begin
        rd_ptr_gray_sync1_reg <= {DEPTH_LOG+1{1'b0}};
        rd_ptr_gray_sync2_reg <= {DEPTH_LOG+1{1'b0}};
    end else begin
        rd_ptr_gray_sync1_reg <= rd_ptr_gray_reg;
        rd_ptr_gray_sync2_reg <= rd_ptr_gray_sync1_reg;
    end
end

always @(posedge m_axis_aclk) begin
    if (output_rst_sync3_reg) begin
        wr_ptr_gray_sync1_reg <= {DEPTH_LOG+1{1'b0}};
        wr_ptr_gray_sync2_reg <= {DEPTH_LOG+1{1'b0}};
    end else begin
        wr_ptr_gray_sync1_reg <= wr_ptr_gray_reg;
        wr_ptr_gray_sync2_reg <= wr_ptr_gray_sync1_reg;
    end
end

// Read logic
always @* begin
    read = 1'b0;

    rd_ptr_next = rd_ptr_reg;
    rd_ptr_gray_next = rd_ptr_gray_reg;

    mem_read_data_valid_next = mem_read_data_valid_reg;

    if (store_output | ~mem_read_data_valid_reg) begin
        // output data not valid OR currently being transferred
        if (~empty) begin
            // not empty, perform read
            read = 1'b1;
            mem_read_data_valid_next = 1'b1;
            rd_ptr_next = rd_ptr_reg + 1;
            rd_ptr_gray_next = rd_ptr_next ^ (rd_ptr_next >> 1);
        end else begin
            // empty, invalidate
            mem_read_data_valid_next = 1'b0;
        end
    end
end

always @(posedge m_axis_aclk) begin
    if (output_rst_sync3_reg) begin
        rd_ptr_reg <= {DEPTH_LOG+1{1'b0}};
        rd_ptr_gray_reg <= {DEPTH_LOG+1{1'b0}};
        mem_read_data_valid_reg <= 1'b0;
    end else begin
        rd_ptr_reg <= rd_ptr_next;
        rd_ptr_gray_reg <= rd_ptr_gray_next;
        mem_read_data_valid_reg <= mem_read_data_valid_next;
    end

    rd_addr_reg <= rd_ptr_next;

    if (read) begin
        mem_read_data_reg <= mem[rd_addr_reg[DEPTH_LOG-1:0]];
    end
end

// Output register
always @* begin
    store_output = 1'b0;

    m_axis_tvalid_next = m_axis_tvalid_reg;

    if (m_axis_tready | ~m_axis_tvalid) begin
        store_output = 1'b1;
        m_axis_tvalid_next = mem_read_data_valid_reg;
    end
end

always @(posedge m_axis_aclk) begin
    if (output_rst_sync3_reg) begin
        m_axis_tvalid_reg <= 1'b0;
    end else begin
        m_axis_tvalid_reg <= m_axis_tvalid_next;
    end

    if (store_output) begin
        output_data_reg <= mem_read_data_reg;
    end
end

endmodule
