module axi4lite_slave
#(
    parameter ADDR_WIDTH = 32,
    parameter DATA_WIDTH = 32//AXI4 spec requires this to be strictly 32 or 64
)
(
//system signals
input axi4l_aclk,
input axi4l_aresetn,//active low, asynchronous assertion and synchronous deassertion

//Write channels
//write address
output reg                  axi4l_awready,
input                       axi4l_awvalid,
input [ADDR_WIDTH-1:0]      axi4l_awaddr,
input [2:0]                 axi4l_awprot,
//write data
output reg                  axi4l_wready,
input                       axi4l_wvalid,
input [DATA_WIDTH-1:0]      axi4l_wdata,
input [(DATA_WIDTH/8)-1:0]  axi4l_wstrb,
//burst response
input                       axi4l_bready,
output reg                  axi4l_bvalid,
output reg [1:0]            axi4l_bresp,//NOTE: 00 = OKAY, 10 = SLVERR (write error)

//Read channels
//read address
output reg                  axi4l_arready,
input                       axi4l_arvalid,
input [ADDR_WIDTH-1:0]      axi4l_araddr,
input [2:0]                 axi4l_arprot,
//read data
input                       axi4l_rready,
output reg                  axi4l_rvalid,
output reg [1:0]            axi4l_rresp,//NOTE: 00 = OKAY, 10 = SLVERR (read error)
output reg [DATA_WIDTH-1:0] axi4l_rdata,

//IP-side interface
input                       ip_clk,
output                      ip_ren,
output                      ip_wen,
output [ADDR_WIDTH-1:0]     ip_addr,
output [(DATA_WIDTH/8)-1:0] ip_wstrb,
output [DATA_WIDTH-1:0]     ip_wdata,
input                       ip_wack,
input                       ip_rack,
input [DATA_WIDTH-1:0]      ip_rdata,
input                       ip_error
);

localparam RESP_OKAY = 2'b00;
localparam RESP_SLVERR = 2'b10;

reg                      internal_ren;
reg                      internal_wen;
reg [ADDR_WIDTH-1:0]     internal_raddr;
reg [ADDR_WIDTH-1:0]     internal_waddr;
reg [(DATA_WIDTH/8)-1:0] internal_wstrb;
reg [DATA_WIDTH-1:0]     internal_wdata;
wire                     internal_wack;
wire                     internal_rack;
wire [DATA_WIDTH-1:0]    internal_rdata;
wire                     internal_error;

wire ip2axi4l_valid, axi4l2ip_valid;
wire candidate_internal_wack;
wire candidate_internal_rack;
wire candidate_internal_error;
wire candidate_internal_ren;
wire candidate_internal_wen;

//check DATA_WIDTH
initial begin
    if(DATA_WIDTH != 32 & DATA_WIDTH != 64) begin
        $display("AXI4Lite DATA_WIDTH must be 32 or 64");
        $finish;
    end
end

//transaction state machine
localparam  STATE_IDLE  = 0,
            STATE_READ  = 1,
            STATE_WRITE = 2;

reg [1:0] state;

always @(posedge axi4l_aclk or negedge axi4l_aresetn)
    if(~axi4l_aresetn)
        state <= STATE_IDLE;
    else case(state)
        STATE_IDLE:
            if(axi4l_awvalid & axi4l_wvalid)
                state <= STATE_WRITE;
            else if(axi4l_arvalid)
                state <= STATE_READ;
        STATE_READ:
            if(axi4l_rvalid & axi4l_rready)
                state <= STATE_IDLE;
        STATE_WRITE:
            if(axi4l_bvalid & axi4l_bready)
                state <= STATE_IDLE;
        default: state <= STATE_IDLE;
    endcase

//write-related internal signals
always @(*) begin
    internal_waddr = axi4l_awaddr;
    internal_wdata = axi4l_wdata;
    internal_wstrb = axi4l_wstrb;
    internal_wen = (state == STATE_IDLE) & axi4l_awvalid & axi4l_wvalid;
end

always @(posedge axi4l_aclk) begin
    axi4l_awready <= internal_wen;
    axi4l_wready <= internal_wen;
end

//read-related internal signals
always @(*) begin
    internal_raddr = axi4l_araddr;
    internal_ren = (state == STATE_IDLE) & ~internal_wen & axi4l_arvalid;
end

always @(posedge axi4l_aclk)
    axi4l_arready <= internal_ren;

axis_async_fifo #(
    .TDATA_SIZE(ADDR_WIDTH+DATA_WIDTH+DATA_WIDTH/8),
    .DEPTH_LOG(2),
    .TUSER_SIZE(2)
) internal_axi4l2ip_fifo (
    .aresetn(axi4l_aresetn),

    .s_axis_aclk(axi4l_aclk),
    .s_axis_tvalid(internal_wen | internal_ren),
    .s_axis_tdata({internal_wen?internal_waddr:internal_raddr,internal_wdata,internal_wstrb}),
    .s_axis_tuser({internal_wen,internal_ren}),
    .s_axis_tready(),//NOTE: FIFO is never full, by design

    .m_axis_aclk(ip_clk),
    .m_axis_tvalid(axi4l2ip_valid),
    .m_axis_tdata({ip_addr,ip_wdata,ip_wstrb}),
    .m_axis_tuser({candidate_ip_wen,candidate_ip_ren}),
    .m_axis_tready(1'b1)
);

assign ip_wen = axi4l2ip_valid & candidate_ip_wen;
assign ip_ren = axi4l2ip_valid & candidate_ip_ren;

//write response on AXI4L bus
always @(posedge axi4l_aclk or negedge axi4l_aresetn)
    if(~axi4l_aresetn) begin
        axi4l_bvalid <= 0;//AXI4 spec requires BVALID pulled LOW during reset
        axi4l_bresp <= RESP_OKAY;
    end else if(internal_wack) begin
        axi4l_bvalid <= 1;
        axi4l_bresp <= internal_error ? RESP_SLVERR : RESP_OKAY;
    end else if(axi4l_bready) begin
        axi4l_bvalid <= 0;
        axi4l_bresp <= RESP_OKAY;
    end

//read response on AXI4L bus
always @(posedge axi4l_aclk or negedge axi4l_aresetn)
    if(~axi4l_aresetn) begin
        axi4l_rvalid <= 0;//AXI4 spec requires RVALID pulled LOW during reset
        axi4l_rdata <= 0;
        axi4l_rresp <= RESP_OKAY;
    end else if(internal_rack) begin
        axi4l_rvalid <= 1;
        axi4l_rdata <= internal_rdata;
        axi4l_rresp <= internal_error ? RESP_SLVERR : RESP_OKAY;
    end else if(axi4l_rready) begin
        axi4l_rvalid <= 0;
        axi4l_rdata <= 0;
        axi4l_rresp <= RESP_OKAY;
    end

axis_async_fifo #(
    .TDATA_SIZE(DATA_WIDTH),
    .DEPTH_LOG(2),
    .TUSER_SIZE(3)
) internal_ip2axi4l_fifo (
    .aresetn(axi4l_aresetn),

    .s_axis_aclk(ip_clk),
    .s_axis_tvalid(ip_wack | ip_rack),
    .s_axis_tdata(ip_rdata),
    .s_axis_tuser({ip_wack,ip_rack,ip_error}),
    .s_axis_tready(),//NOTE: FIFO is never full, by design, so it is always ready

    .m_axis_aclk(axi4l_aclk),
    .m_axis_tvalid(ip2axi4l_valid),
    .m_axis_tdata(internal_rdata),
    .m_axis_tuser({candidate_internal_wack,candidate_internal_rack,candidate_internal_error}),
    .m_axis_tready(1'b1)
);

assign internal_wack = ip2axi4l_valid & candidate_internal_wack;
assign internal_rack = ip2axi4l_valid & candidate_internal_rack;
assign internal_error = candidate_internal_error | (candidate_internal_wack & candidate_internal_rack);

endmodule
