module axi_hacm #(
    parameter ADDR_WIDTH = 10,
    parameter DATA_WIDTH = 32
) (
    // AXI4-Lite Interface
    input                    axi4l_aclk,
    input                    axi4l_aresetn,
    // Write address channel
    input                    axi4l_awvalid,
    output                   axi4l_awready,
    input   [ADDR_WIDTH-1:0] axi4l_awaddr,
    input              [2:0] axi4l_awprot,
    // Write data channel
    input                    axi4l_wvalid,
    output                   axi4l_wready,
    input [DATA_WIDTH  -1:0] axi4l_wdata,
    input [DATA_WIDTH/8-1:0] axi4l_wstrb,
    // Write response channel
    output                   axi4l_bvalid,
    input                    axi4l_bready,
    output             [1:0] axi4l_bresp, // Note: 00 = OKAY, 10 = SLVERR (write error)
    // Read address channel
    input                    axi4l_arvalid,
    output                   axi4l_arready,
    input   [ADDR_WIDTH-1:0] axi4l_araddr,
    input              [2:0] axi4l_arprot,
    // Read data channel
    output                   axi4l_rvalid,
    input                    axi4l_rready,
    output             [1:0] axi4l_rresp, // Note: 00 = OKAY, 10 = SLVERR (read error)
    output  [DATA_WIDTH-1:0] axi4l_rdata,

    // HACM Inputs
    input  f_or_s_i,

    // HACM Outputs
    output f_or_s_o,
    output lock_o,
    output clk_o
);

    // Configuration interface
    wire        reg_clk;
    wire        reg_rden;
    wire        reg_wren;
    wire [ 3:0] reg_wstrb;
    wire [31:0] reg_rd_data;
    wire [31:0] reg_wr_data;
    wire [ADDR_WIDTH-1:0] reg_addr;
    reg         reg_wack;
    reg         reg_rack;

    axi4lite_slave #(
        .ADDR_WIDTH(ADDR_WIDTH),
        .DATA_WIDTH(DATA_WIDTH)
    ) axi4l_slave (
        .axi4l_aclk   (axi4l_aclk   ),
        .axi4l_aresetn(axi4l_aresetn),
        // Write address channel
        .axi4l_awvalid(axi4l_awvalid),
        .axi4l_awready(axi4l_awready),
        .axi4l_awaddr (axi4l_awaddr ),
        .axi4l_awprot (axi4l_awprot ),
        // Write data channel
        .axi4l_wvalid (axi4l_wvalid ),
        .axi4l_wready (axi4l_wready ),
        .axi4l_wdata  (axi4l_wdata  ),
        .axi4l_wstrb  (axi4l_wstrb  ),
        // Write response channel
        .axi4l_bvalid (axi4l_bvalid ),
        .axi4l_bready (axi4l_bready ),
        .axi4l_bresp  (axi4l_bresp  ),
        // Read address channel
        .axi4l_arvalid(axi4l_arvalid),
        .axi4l_arready(axi4l_arready),
        .axi4l_araddr (axi4l_araddr ),
        .axi4l_arprot (axi4l_arprot ),
        // Read data channel
        .axi4l_rvalid (axi4l_rvalid ),
        .axi4l_rready (axi4l_rready ),
        .axi4l_rresp  (axi4l_rresp  ),
        .axi4l_rdata  (axi4l_rdata  ),

        .ip_clk  (reg_clk),
        .ip_ren  (reg_rden),
        .ip_wen  (reg_wren),
        .ip_addr (reg_addr),
        .ip_wstrb(reg_wstrb),
        .ip_wdata(reg_wr_data),
        .ip_wack (reg_wack),
        .ip_rack (reg_rack),
        .ip_rdata(reg_rd_data),
        .ip_error(1'b0)
    );

    hacm hacm_inst (
        .clk_i      ( axi4l_aclk   ),
        .rst_i      (~axi4l_aresetn),
        // The following inputs are synchronous to clk_o
        .f_or_s_i   (f_or_s_i      ),
        .lock_o     (lock_o        ),
        .clk_o      (clk_o         ),
        // Configuration interface
        .reg_clk    (reg_clk       ),
        .reg_rden   (reg_rden      ),
        .reg_wren   (reg_wren      ),
        .reg_wstrb  (reg_wstrb     ),
        .reg_rd_data(reg_rd_data   ),
        .reg_wr_data(reg_wr_data   ),
        .reg_addr   (reg_addr      )
    );

    always @(posedge axi4l_aclk or negedge axi4l_aresetn)
        if (!axi4l_aresetn) begin
            reg_wack <= 0;
            reg_rack <= 0;
        end
        else begin
            reg_wack <= reg_wren;
            reg_rack <= reg_rden;
        end

    assign reg_clk = axi4l_aclk;

    assign f_or_s_o = f_or_s_i;
endmodule
