-------------------------------------------------------------------------------
-- Project     : HACM
-- Author      : Alexandru Gheolbanoiu, Lucian Petrica
-- Date        : 31.08.2014
-- File        : hacm_control.vhd
-- Description : 
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
library unisim;
use unisim.vcomponents.all;

-- ENTITY ---------------------------------------------------------------------
-- the HACM control unit receives timing and instruction info and
-- decides which ACM strategy to utilize (stretching or multiplexing)
entity hacm_control is
	port 
	(
		clk_i	    : in  std_logic;
		rst_i     : in  std_logic;
		
        --fixed-point, 5.3 format, value in nanoseconds
		f_or_s_i  : in  std_logic;                     --indicates fast or slow instructions (1=slow, 0=fast)
		ts_i      : in  std_logic_vector(7 downto 0);  --period of slow clock (desired)
		tf_i      : in  std_logic_vector(7 downto 0);  --period of fast clock
		to_i      : in  std_logic_vector(7 downto 0);  --period of slow clock (attainable)
		tsw_i     : in  std_logic_vector(7 downto 0);  --duration of a clock switch
		
		mode_o    : out std_logic;
		stretch_o : out std_logic_vector(1 downto 0);	
		sel_o     : out std_logic
	);
end hacm_control;

-- ARCHITECTURE ---------------------------------------------------------------
architecture rtl of hacm_control is

  --CM/CS Decision
  signal sh_reg_s : std_logic_vector(99 downto 0);
  signal nsi_reg_s: std_logic_vector( 6 downto 0);
  signal nfi_reg_s: std_logic_vector( 6 downto 0);
  signal ncs_reg_s: std_logic_vector( 6 downto 0);  
  signal tcm  : std_logic_vector(19 downto 0);
  signal tcs  : std_logic_vector(19 downto 0);
  signal tsi  : std_logic_vector(14 downto 0);
  signal tswi : std_logic_vector(14 downto 0);
  signal tscs : std_logic_vector(14 downto 0);  
  signal tsi_r  : std_logic_vector(14 downto 0);
  signal tswi_r : std_logic_vector(14 downto 0);
  signal tscs_r : std_logic_vector(14 downto 0);  
  signal mode_reg_s : std_logic;
  
  --CS/CM MUX CTRL
  
  
begin


--f_or_s shift register
  process(clk_i)
  begin
    if(clk_i'event and clk_i = '1') then
        sh_reg_s(99 downto 1) <= sh_reg_s(98 downto 0);
        if(rst_i = '1') then
            sh_reg_s(0) <= '0';
        else
            sh_reg_s(0) <= f_or_s_i;
        end if;
    end if;
  end process;

--CM/CS Decision
  process(clk_i)
  begin
    if(clk_i'event and clk_i = '1') then 	 
      if(rst_i = '1') then 
		  nsi_reg_s <= (others => '0');
		  nfi_reg_s <= "1100011";
		  ncs_reg_s <= (others => '0');
		  
		  tsi    <= (others => '0');
	     tswi   <= (others => '0');
	     tscs   <= (others => '0');
	     tsi_r  <= (others => '0');
	     tswi_r <= (others => '0');
	     tscs_r <= (others => '0');
        tcm    <= (others => '0');
	     tcs    <= (others => '0');
	     mode_reg_s <= '0';
      else
        nsi_reg_s <= nsi_reg_s + (f_or_s_i & nand_reduce(nsi_reg_s)) - (sh_reg_s(99) & or_reduce(nsi_reg_s)); 
		  nfi_reg_s <= nfi_reg_s + (not(f_or_s_i) & nand_reduce(nfi_reg_s)) - (not(sh_reg_s(99)) & or_reduce(nfi_reg_s));  
		  ncs_reg_s <= ncs_reg_s + (f_or_s_i xor sh_reg_s(0)) - (sh_reg_s(99) xor sh_reg_s(98));

        		 
        tsi    <= nsi_reg_s*ts_i;
	     tswi   <= ncs_reg_s*tsw_i;
	     tscs   <= nsi_reg_s*to_i;
	     tsi_r  <= tsi;
	     tswi_r <= tswi;
	     tscs_r <= tscs;
        tcm    <= "0000" & (('0' & tsi_r) + ('0' & tswi_r));
	     tcs    <= "00000" & tscs_r;
		  if(tcs > tcm) then mode_reg_s <= '1'; else mode_reg_s <= '0'; end if;			 
		end if;	
	 end if;
  end process; 
  mode_o <= mode_reg_s;
  
--CS/CM MUX CTRL
  stretch_o <= "01" when f_or_s_i = '1' else "00";
  sel_o     <= f_or_s_i;

-------------------------------------------------------------------------------
end rtl;